{*************************************************************************}
{*   Storage                                                             *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Data\Storage.pas                                                    *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 17th February 2003  (17.02.2003)                           }
{*************************************************************************}


{This unit contains base data storage/access classes,
types, functions and interfaces.}


unit Storage;


{TODO -o Galfar : Add Vampyre File Format - base for all our files}
{DONE -o Galfar : repair RLE. For some files adds one byte at the end.}
{DONE -o Galfar : Redesign FormatCodes - they are not file format magics, but
 are from module Identification - FormatCode = 'IPNG' indicates that
 module works with images in png format, ....}
{DONE -o Galfar : Add to all XXX.Load/SaveFromFile code
to catch exception EOpenFileError atd. Look at TFileStream implementation}


interface


uses
  Sys, Objects, SysUtils;


type


  EStorage = class(Exception);

  ECompress = class(EStorage);


{*************************************************************************}
{                        Storage Modules                                  }
{*************************************************************************}


  {Interface for TStorageModule.}
  IStorageModule = interface(IModule)
    ['{F721681B-DC98-4A93-87DF-649249BFB168}']

  end;


  {Base class for all modules used in Data sources}
  TStorageModule = class(TModule, IStorageModule)
  protected

  public

    constructor Create; override;
    destructor Destroy; override;


  published

  end;


  {TStorageModule Class reference}
  TStorageClass = class of TStorageModule;


  {Interface for TCompressModule}
  ICompressModule = interface(IStorageModule)
    ['{2354AFFC-3113-46F2-8CBF-9677D3FC1768}']

    function GetCanCompress: boolean;
    function GetCanDecompress: boolean;

    function CompressBuffer(InData: Pointer; InSize: Long; var OutData: Pointer;
      var OutSize: Long): boolean;
    function CompressStream(Input: TMemoryStream): boolean;
    function CompressFile(const InFile, OutFile: string): boolean;
    function DecompressStream(Input: TMemoryStream): boolean;
    function DecompressBuffer(InData: Pointer; InSize: Long; var OutData:
      Pointer; var OutSize: Long): boolean;
    function DecompressFile(const InFile, OutFile: string): boolean;

    property CanCompress: boolean read GetCanCompress;
    property CanDecompress: boolean read GetCanDecompress;

  end;


const

  {ID Code for descendants of TCompressModule is 'CMP' so modules can
  have IDcodes like: CMP-RLE, CMP-BZ2, ...}
  CompressModuleID = 'CMP';


type

  {Base class for all compression/decomression modules like ZLIB, BZ2, LHZ.
  Compression modules are not for opening archives like ZIP, they are
  only for memory pack/unpack operations. However, some archivers/imageloaders
  should  use some CompressModules for decoding contained data.
  FormatCodes are not used directly for but are used in custom
  file formats for identifying with which compressor was certain file
  block packed.}
  TCompressModule = class(TStorageModule, ICompressModule)
  protected

    FCanCompress: boolean;
    FCanDecompress: boolean;

    function GetCanCompress: boolean;
    function GetCanDecompress: boolean;

  public

    constructor Create; override;
    destructor Destroy; override;

    {Compresses InData buffer and returns compressed data in OutData and
     size in OutSize. Returns true if action was succesful.}
    function CompressBuffer(InData: Pointer; InSize: Long; var OutData: Pointer;
      var OutSize: Long): boolean; virtual;

    {Compresses entire input stream. Compressed data are written back to input stream.
     Returns true if action was succesful. Do not use when you want to
     compress only part of stream !!}
    function CompressStream(Input: TMemoryStream): boolean; virtual;

    {Compresses input file to output file.Returns true if action was succesful.}
    function CompressFile(const InFile, OutFile: string): boolean; virtual;

    {Decompresses InData buffer and returns decompressed data in OutData and
    size in OutSize. Returns true if action was succesful.}
    function DecompressBuffer(InData: Pointer; InSize: Long; var OutData:
      Pointer; var OutSize: Long): boolean; virtual;

    {Decompresses entire(!) input stream. Decompressed data are written back to input stream.
     Returns true if action was succesful. Do not use when you want to
     decompress only part of stream !!}
    function DecompressStream(Input: TMemoryStream): boolean; virtual;

    {Deompresses input file to output file.Returns true if action was succesful.}
    function DecompressFile(const InFile, OutFile: string): boolean; virtual;


    {Indicates if module can compress data}
    property CanCompress: boolean read GetCanCompress;

    {Indicates if module can decompress data (It is true in "all" modules}
    property CanDecompress: boolean read GetCanDecompress;


  published

  end;


  {TCompressModule Class reference}
  TCompressClass = class of TCompressModule;


{*************************************************************************}
{                       External Storage Module Support                   }
{*************************************************************************}


  {Remember that these functions must be given allocated
  OutData when called. They can not call GetMem(OutData) because
  later it will cause Error when you try to Free this memory
  (it belongs to another process). Use simple GetMem(OutData,1) before call.
  If you are using TExtCompressModule it will be done automaticaly for you.}
 
  {Compress buffer function type}
  TCompressBufferProc = function(InData: Pointer; InSize: Long; var OutData:
    Pointer; var OutSize: Long): boolean; cdecl;


  {Decompress buffer function type}
  TDecompressBufferProc = function(InData: Pointer; InSize: Long; var OutData:
    Pointer; var OutSize: Long): boolean; cdecl;


  {This record is returned from Module Library (dll/bpl/so) to TExtCompressModule
  and this class then assigns returned functions to self functions .}
  TExtCompressInfo = packed record
    CanCompress: boolean;
    CanDecompress: boolean;
    CompressBuffer: TCompressBufferProc;
    DecompressBuffer: TDecompressBufferProc;
  end;

  PExtCompressInfo = ^TExtCompressInfo;


  {Module wrapper for low-level Compress modules retrieved from
  External Module Libraries.
  External low-level TCompressModule must implement CompressBuffer/DecompressBuffer
  functions (other are based on these two).}
  TExtCompressModule = class(TCompressModule, IExternalModule)
  protected

    FExtInfo: TExtCompressInfo;

  public

    constructor Create; override;
    destructor Destroy; override;

    {In this function all data and functions for given module are
    loaded from library defined by Handle. Index identifies module
    in multi-module libraries if GetModuleCount(Handle)>1.
    Returns true if Module succesfully loaded props/funcs from
    library.}
    function LoadFromLib(Handle: THandle; Index: Long = 0): boolean;

    {Compresses InData buffer and returns compressed data in OutData and
    size in OutSize. Returns true if action was succesful.}
    function CompressBuffer(InData: Pointer; InSize: Long; var OutData: Pointer;
      var OutSize: Long): boolean; override;

    {Decompresses InData buffer and returns decompressed data in OutData and
    size in OutSize. Returns true if action was succesful.}
    function DecompressBuffer(InData: Pointer; InSize: Long; var OutData:
      Pointer; var OutSize: Long): boolean; override;


  published

  end;


const

  sExtCmpName = 'Compress Module Wrapper';
  sExtCmpDesc = 'Provides access to low-level external compression module';
  sExtCmpCode = 'CMP-EXT';


{*************************************************************************}
{                         Storage Engine                                 }
{*************************************************************************}


type

  {Interface for TStorageEngine}
  IStorageEngine = interface(IModEngine)
    ['{F78CCB16-DEB6-4530-8F95-BFF912A139DA}']

    function GetActiveCompress: ICompressModule;
    function GetCompressModule(Index: Long): ICompressModule;
    function GetCompressModuleCount: Long;
    procedure SetActiveCompress(Value: ICompressModule);

    function CompressBuffer(InData: Pointer; InSize: Long; out OutData: Pointer;
      out OutSize: Long): boolean;
    function CompressStream(Input: TMemoryStream): boolean;
    function CompressFile(const InFile, OutFile: string): boolean;
    function DecompressBuffer(InData: Pointer; InSize: Long; out OutData:
      Pointer;
      out OutSize: Long): boolean;
    function DecompressStream(Input: TMemoryStream): boolean;
    function DecompressFile(const InFile, OutFile: string): boolean;

    property ActiveCompress: ICompressModule read GetActiveCompress write
      SetActiveCompress;
    property CompressModule[Index: Long]: ICompressModule read
      GetCompressModule;
    property CompressModuleCount: Long read GetCompressModuleCount;

  end;


  {Global storage engine which manages TCompressModules and dooes
  many other things.}
  TStorageEngine = class(TModEngine, IStorageEngine)
  protected

    function GetActiveCompress: ICompressModule;
    function GetCompressModule(Index: Long): ICompressModule;
    function GetCompressModuleCount: Long;

    procedure SetActiveCompress(Value: ICompressModule);

  public

    constructor Create; override;
    destructor Destroy; override;


    {These functions class samelly named functions of
    contained CompressModules. First ActiveCompress call is tried
    and if it returns True it is all right. When it returns False
    all CompressModules with same IDCode are tried. If none
    of then is successful Error is raised.
    See details on these functions in TCompressModule declaration.}
    function CompressBuffer(InData: Pointer; InSize: Long; out OutData: Pointer;
      out OutSize: Long): boolean;
    function CompressStream(Input: TMemoryStream): boolean;
    function CompressFile(const InFile, OutFile: string): boolean;
    function DecompressBuffer(InData: Pointer; InSize: Long; out OutData:
      Pointer;
      out OutSize: Long): boolean;
    function DecompressStream(Input: TMemoryStream): boolean;
    function DecompressFile(const InFile, OutFile: string): boolean;


    {Active module used for compress/decompress operations.}
    property ActiveCompress: ICompressModule read GetActiveCompress write
      SetActiveCompress;
    {Returns compress module by given index.}
    property CompressModule[Index: Long]: ICompressModule read
      GetCompressModule;
    {Number of compress modules.}
    property CompressModuleCount: Long read GetCompressModuleCount;


  published

  end;


{*************************************************************************}
{                         Storage Classes                                 }
{*************************************************************************}


  {Use this class as base class for all objects which stores some
  form of data (images, buffers, ....)}
  TStorageObject = class(TBaseObject)
  protected

  public

    constructor Create; override;
    destructor Destroy; override;

    {Returns size of memory in bytes occupied by data stored in
    this object.}
    function MemoryUsage: Long; virtual;


  published

  end;


const

  sStorageEngineName = 'Storage Engine';
  sStorageEngineDesc = 'Manages compression modules and provides access ' +
    'to their functionality';


{*************************************************************************}
{                         Misc functions                                  }
{*************************************************************************}


{This function calculates CRC32 vaule for given buffer.}
function CalcCRC32(Data: Pointer; Size: Long): ULong; overload;

{This procedure calculates CRC32 vaule for given buffer. Starting CRC32 val
can be set here.}
procedure CalcCRC32(Data: Pointer; Size: Long; var CRC32: ULong); overload;


{*************************************************************************}
{                      Some other types/const                             }
{*************************************************************************}


type

  {Table used in CRC32 calculations.}
  TCRC32Table = array[Byte] of ULong;


var

  {Interface to StorageEngineInstance}
  StorageEngine: IStorageEngine;


implementation


uses
  {Internal storage modules that are automatically registered}
  RleModule, ZLibModule, BZip2Module;


var

  {Instance of TStorageEngine}
  StorageEngineInstance: TStorageEngine;

  {Module Info for TExtCompressModule}
  ExtCmpModuleInfo: TModuleInfo;


{*************************************************************************}
{                         TStorageModule Methods                          }
{*************************************************************************}


constructor TStorageModule.Create;
begin
  inherited Create;
end;

destructor TStorageModule.Destroy;
begin
  inherited Destroy;
end;


{*************************************************************************}
{                         TCompressModule Methods                         }
{*************************************************************************}


constructor TCompressModule.Create;
begin
  inherited Create;
  FCanCompress := true;
  FCanDecompress := true;
end;

destructor TCompressModule.Destroy;
begin

  inherited Destroy;
end;

function TCompressModule.GetCanCompress: boolean;
begin
  Result := FCanCompress;
end;

function TCompressModule.GetCanDecompress: boolean;
begin
  Result := FCanDecompress;
end;

function TCompressModule.CompressBuffer(InData: Pointer;
  InSize: Long; var OutData: Pointer; var OutSize: Long): boolean;
begin
  Result := false;
end;

function TCompressModule.CompressStream(Input: TMemoryStream): boolean;
var
  Data: Pointer;
  Size: Long;
begin
  Data := nil;
  Result := false;
  Assert(Input <> nil, ErrString(Self, 'CoompressStream',
    'Input stream not created! What should I compress then?'));
  try
    try
      Result := CompressBuffer(Input.Memory, Input.Size, Data, Size);
      if Result then
      begin
        Input.Clear;
        Input.WriteBuffer(Data^, Size);
        Input.Position := 0;
      end;
    except
      RaiseError(ErrString(Self, 'CompressStream',
        'Error when compressing stream!'), ECompress);
    end;
  finally
    FreeMem(Data);
  end;
end;

function TCompressModule.CompressFile(const InFile,
  OutFile: string): boolean;
var
  FI, FO: TFileStream;
  MS: TMemoryStream;

begin
  Result := false;
  MS := TMemoryStream.Create;
  FI := nil;
  FO := nil;
  try
    try
      FI := TFileStream.Create(InFile, fmOpenRead);
      FO := TFileStream.Create(OutFile, fmCreate);
    except
      on E2: EFOpenError do
        RaiseError(ErrString(Self, 'CompressFile', E2.Message), EFOpenError);
      on E1: EFCreateError do
        RaiseError(ErrString(Self, 'CompressFile', E1.Message), EFCreateError);
    end;
    try
      MS.CopyFrom(FI, FI.Size);
      Result := CompressStream(MS);
      FO.CopyFrom(MS, MS.Size);
    except
      RaiseError(ErrString(Self, 'CompressFile', 'Error when compressing file "'
        + InFile + '" to "' + OutFile + '" !'), ECompress);
    end;
  finally
    FI.Free;
    FO.Free;
    MS.Free;
  end;
end;

function TCompressModule.DecompressBuffer(InData: Pointer;
  InSize: Long; var OutData: Pointer; var OutSize: Long): boolean;
begin
  Result := false;
end;

function TCompressModule.DecompressStream(Input: TMemoryStream): boolean;
var
  Data: Pointer;
  Size: Long;
begin
  Data := nil;
  Result := false;
  Assert(Input <> nil, ErrString(Self, 'DecompressStream',
    'Input stream not created! What should I decompress then?'));
  try
    try
      Input.Position := 0;
      Result := DecompressBuffer(Input.Memory, Input.Size, Data, Size);
      if Result then
      begin
        Input.Clear;
        Input.WriteBuffer(Data^, Size);
        Input.Position := 0;
      end;
    except
      RaiseError(ErrString(Self, 'DecompressStream',
        'Error when decompressing stream!' + sLineBreak +
        'Data can be corrupted or uncompressed!'), ECompress);
    end;
  finally
    FreeMem(Data);
  end;
end;

function TCompressModule.DecompressFile(const InFile,
  OutFile: string): boolean;
var
  FI, FO: TFileStream;
  MS: TMemoryStream;

begin
  Result := false;
  MS := TMemoryStream.Create;
  FI := nil;
  FO := nil;
  try
    try
      FI := TFileStream.Create(InFile, fmOpenRead);
      FO := TFileStream.Create(OutFile, fmCreate);
    except
      on E2: EFOpenError do
        RaiseError(ErrString(Self, 'DecompressFile', E2.Message), EFOpenError);
      on E1: EFCreateError do
        RaiseError(ErrString(Self, 'DecompressFile', E1.Message),
          EFCreateError);
    end;
    try
      MS.CopyFrom(FI, FI.Size);
      Result := DecompressStream(MS);
      FO.CopyFrom(MS, MS.Size);
    except
      RaiseError(ErrString(Self, 'DecompressFile',
        'Error when decompressing file "'
        + InFile + '" to "' + OutFile + '" !'), ECompress);
    end;
  finally
    FI.Free;
    FO.Free;
    MS.Free;
  end;
end;


{*************************************************************************}
{                       External Storage Module Support                   }
{*************************************************************************}


constructor TExtCompressModule.Create;
begin
  inherited Create;
  FInfo := ExtCmpModuleInfo;
end;

destructor TExtCompressModule.Destroy;
begin

  inherited Destroy;
end;

function TExtCompressModule.LoadFromLib(Handle: THandle;
  Index: Long): boolean;
var
  P: Pointer;

begin
  Result := false;
  if not IsValidModuleLib(Handle) then
  begin
    RaiseError(ErrString(Self, 'LoadFromLib', 'Library "' +
      GetLibName(Handle) + '" is not valid External Module Library!'),
      ECompress);
    Exit;
  end;

  try
    FInfo := Objects.GetModuleInfo(Handle, Index);
  except
    RaiseError(ErrString(Self, 'LoadFromLibe', 'Error when loading'
      + ' properties and functions of external module from library "' +
      GetLibName(Handle) + '"!'), ECompress);
  end;

  P := Objects.GetModuleExtInfo(Handle, Index);
  Assert(P <> nil, ErrString(Self, 'LoadFromLib',
    'GetModuleExtInfo returned nil.' +
    ' Module "' + FInfo.Name + '" from library "' + GetLibName(Handle) +
    '" is badly written!'));

  try
    // assign properties
    FExtInfo := PExtCompressInfo(P)^;

    FName := FInfo.Name;
    FDescription := FInfo.Description;

    FCanCompress := FExtInfo.CanCompress;
    FCanDecompress := FExtInfo.CanDecompress;

  except
    RaiseError(ErrString(Self, 'LoadFromLib',
      'Error when assigning' +
      ' properties loaded from external module from library "' +
      GetLibName(Handle) + '"!'), ECompress);
  end;
end;

function TExtCompressModule.CompressBuffer(InData: Pointer; InSize: Long;
  var OutData: Pointer;var OutSize: Long): boolean;
begin
  Result := false;
  if (@FExtInfo.CompressBuffer <> nil) and FExtInfo.CanCompress then
  try
    Result := FExtInfo.CompressBuffer(InData, InSize, OutData, OutSize);
  except
    RaiseError(ErrString(Self, 'CompressBuffer',
      'Error when executing' +
      ' CompressBuffer from external library!'), ECompress);
    FreeMemNil(OutData);
  end
  else
    RaiseError(ErrString(Self, 'CompressBuffer',
      'This external module' +
      ' does not support CompressBuffer function!'), ECompress);
end;

function TExtCompressModule.DecompressBuffer(InData: Pointer; InSize: Long;
  var OutData: Pointer; var OutSize: Long): boolean;
begin
  Result := false;
  if (@FExtInfo.DecompressBuffer <> nil) and FExtInfo.CanDecompress then
  try
    Result := FExtInfo.DecompressBuffer(InData, InSize, OutData, OutSize);
  except
    RaiseError(ErrString(Self, 'DecompressBuffer',
      'Error when executing' +
      ' DecompressBuffer from external library!'), ECompress);
    FreeMemNil(OutData);
  end
  else
    RaiseError(ErrString(Self, 'DecompressBuffer',
      'This external module' +
      ' does not support DecompressBuffer function!'), ECompress);
end;


{*************************************************************************}
{                      TStorageEngine Methods                             }
{*************************************************************************}


constructor TStorageEngine.Create;
begin
  inherited Create;
  // FRefLists[0] contains CompressModules
  FSupportedCodes.Add(CompressModuleID);

end;

destructor TStorageEngine.Destroy;
begin

  inherited Destroy;
end;

function TStorageEngine.GetActiveCompress: ICompressModule;
begin
  Result := ICompressModule(FActiveModules[0]);
end;

function TStorageEngine.GetCompressModule(Index: Long): ICompressModule;
begin
  Result := ICompressModule(FRefLists[0][Index]);
end;

function TStorageEngine.GetCompressModuleCount: Long;
begin
  Result := FRefLists[0].Count;
end;

procedure TStorageEngine.SetActiveCompress(Value: ICompressModule);
begin
  if Value <> nil then
    FActiveModules[0] := Pointer(Value);
end;

function TStorageEngine.CompressBuffer(InData: Pointer; InSize: Long; out
  OutData: Pointer;
  out OutSize: Long): boolean;
var
  I: Long;
begin
  Result := false;
  OutData := nil;
  OutSize := 0;
  if ActiveCompress = nil then
    Exit;

  Result := ActiveCompress.CompressBuffer(InData, InSize, OutData, OutSize);
  I := 0;
  while (not Result) and (I < CompressModuleCount) do
  begin
  // use only same IDCode but not ActiveCompress again
    if (CompressModule[I].ModuleInfo.IDCode =
      ActiveCompress.ModuleInfo.IDCode) and
      (CompressModule[I].ModuleInfo.Name <> ActiveCompress.ModuleInfo.Name) then

      Result := CompressModule[I].CompressBuffer(InData, InSize, OutData,
        OutSize);
    Inc(I);
  end;
end;

function TStorageEngine.CompressStream(Input: TMemoryStream): boolean;
var
  I: Long;
begin
  Result := false;
  if ActiveCompress = nil then
    Exit;

  Result := ActiveCompress.CompressStream(Input);
  I := 0;
  while (not Result) and (I < CompressModuleCount) do
  begin
  // use only same IDCode but not ActiveCompress again
    if (CompressModule[I].ModuleInfo.IDCode =
      ActiveCompress.ModuleInfo.IDCode) and
      (CompressModule[I].ModuleInfo.Name <> ActiveCompress.ModuleInfo.Name) then

      Result := CompressModule[I].CompressStream(Input);
    Inc(I);
  end;
end;

function TStorageEngine.CompressFile(const InFile,
  OutFile: string): boolean;
var
  I: Long;
begin
  Result := false;
  if ActiveCompress = nil then
    Exit;
  Result := ActiveCompress.CompressFile(InFile, OutFile);
  I := 0;
  while (not Result) and (I < CompressModuleCount) do
  begin
  // use only same IDCode but not ActiveCompress again
    if (CompressModule[I].ModuleInfo.IDCode =
      ActiveCompress.ModuleInfo.IDCode) and
      (CompressModule[I].ModuleInfo.Name <> ActiveCompress.ModuleInfo.Name) then

      Result := CompressModule[I].CompressFile(InFile, OutFile);
    Inc(I);
  end;
end;

function TStorageEngine.DecompressBuffer(InData: Pointer; InSize: Long; out
  OutData: Pointer;
  out OutSize: Long): boolean;
var
  I: Long;
begin
  Result := false;
  OutData := nil;
  OutSize := 0;
  if ActiveCompress = nil then
    Exit;

  Result := ActiveCompress.DecompressBuffer(InData, InSize, OutData, OutSize);
  I := 0;
  while (not Result) and (I < CompressModuleCount) do
  begin
  // use only same IDCode but not ActiveCompress again
    if (CompressModule[I].ModuleInfo.IDCode =
      ActiveCompress.ModuleInfo.IDCode) and
      (CompressModule[I].ModuleInfo.Name <> ActiveCompress.ModuleInfo.Name) then

      Result := CompressModule[I].DecompressBuffer(InData, InSize, OutData,
        OutSize);
    Inc(I);
  end;
end;

function TStorageEngine.DecompressStream(Input: TMemoryStream): boolean;
var
  I: Long;
begin
  Result := false;
  if ActiveCompress = nil then
    Exit;

  Result := ActiveCompress.DecompressStream(Input);
  I := 0;
  while (not Result) and (I < CompressModuleCount) do
  begin
  // use only same IDCode but not ActiveCompress again
    if (CompressModule[I].ModuleInfo.IDCode =
      ActiveCompress.ModuleInfo.IDCode) and
      (CompressModule[I].ModuleInfo.Name <> ActiveCompress.ModuleInfo.Name) then

      Result := CompressModule[I].DecompressStream(Input);
    Inc(I);
  end;
end;

function TStorageEngine.DecompressFile(const InFile,
  OutFile: string): boolean;
var
  I: Long;
begin
  Result := false;
  if ActiveCompress = nil then
    Exit;
  Result := ActiveCompress.DecompressFile(InFile, OutFile);
  I := 0;
  while (not Result) and (I < CompressModuleCount) do
  begin
  // use only same IDCode but not ActiveCompress again
    if (CompressModule[I].ModuleInfo.IDCode =
      ActiveCompress.ModuleInfo.IDCode) and
      (CompressModule[I].ModuleInfo.Name <> ActiveCompress.ModuleInfo.Name) then

      Result := CompressModule[I].DecompressFile(InFile, OutFile);
    Inc(I);
  end;
end;


{*************************************************************************}
{                         Storage Classes                                 }
{*************************************************************************}


constructor TStorageObject.Create;
begin
  inherited Create;

end;

destructor TStorageObject.Destroy;
begin

  inherited Destroy;
end;

function TStorageObject.MemoryUsage: Long;
begin
  Result := 0;
end;


{*************************************************************************}
{                         Misc functions                                  }
{*************************************************************************}


var
  CRC32Table: TCRC32Table;

procedure InitCRC32Table;
const
  Polynom = $EDB88320;
var
  I, J: Long;
  C: ULong;
begin
  for I := 0 to 255 do
  begin
    C := I;
    for J := 0 to 7 do
    begin
      if (C and $01) <> 0 then
        C := Polynom xor (C shr 1)
      else
        C := C shr 1;
    end;
    CRC32Table[I] := C;
  end;
end;

function CalcCRC32(Data: Pointer; Size: Long): ULong;
var
  I: Long;
  B: PByte;
begin
  Result := $FFFFFFFF;
  B := Data;
  for I := 0 to Size - 1 do
  begin
    Result := (Result shr 8) xor
      CRC32Table[B^ xor Byte(Result)];
    Inc(B);
  end
end;

procedure CalcCRC32(Data: Pointer; Size: Long; var CRC32: ULong);
var
  I: Long;
  B: PByte;
begin
  B := Data;
  for I := 0 to Size - 1 do
  begin
    CRC32 := (CRC32 shr 8) xor
      CRC32Table[B^ xor Byte(CRC32)];
    Inc(B);
  end
end;


{*************************************************************************}
{                       Some impostant functions                          }
{*************************************************************************}


procedure InitStorageEngine;
begin
  StorageEngineInstance := TStorageEngine.Create;
  StorageEngine := StorageEngineInstance;
end;

procedure FreeStorageEngine;
begin
  ReleaseIntf(StorageEngine);
  StorageEngineInstance := nil;
end;



initialization

  // Create  CRC32 look up table here
  InitCRC32Table;

  InitStorageEngine;

  // Registers TExtCompressModule to be usable as wrapper to
  // external low-level compression modules
  ExtCmpModuleInfo := GetDefaultIntModuleInfo;
  ExtCmpModuleInfo.Name := sExtCmpName;
  ExtCmpModuleInfo.Description := sExtCmpDesc;
  ExtCmpModuleInfo.IDCode := sExtCmpCode;
  ExtCmpModuleInfo.Style := msAbsWrapper;
  RegisterIntModule(@ExtCmpModuleInfo, TExtCompressModule);

finalization

  FreeStorageEngine;

end.

