{*************************************************************************}
{*   PngModule                                                           *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Graphic\Images\Png\PngModule.pas                                    *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 25th February 2003  (25.02.2003)                           }
{*************************************************************************}


{This unit contains module for loading/saving PNG images.}


unit PngModule;


{TODO -o Galfar : Images with multiple IDATs may not work!}
{DONE -o Galfar : Repair interlace loading}
{DONE -o Galfar : Saved images can not be opened in some progs (ACDSee) }
{DONE -o Galfar : Move SwapLOng to Sys.pas}


interface


uses
  Sys, Storage, Images, Objects, SysUtils, Math,
  {Uses ZLib for image data compression/decompression}
  ZlibEx;


type

  {Image encoder for loading/saving png file format.
  Supports: 1,4,8,16,24,32bpp images(filtered/noninterlaced/interlaced) loading
            8,16,24,32 noninterlaced nonfiltered saving.
  Compression: ZLIB.}
  TPngModule = class(TImageModule)
  protected

    FStream: TZStreamRec;

    procedure ApplyFilter(Filter: Byte; Line, PrevLine, Target: PByte;
      BPP, BytesPerRow: Integer);

    {Here convert from image formats not supported by encoder }
    procedure PrepareImage(const Image: TImageData; out Clone: TImageData);
      override;


  public

    constructor Create; override;
    destructor Destroy; override;

    {This function converts ImageData structure to png.}
    function EncodeBuffer(const Image: TImageData; out Data: Pointer; out Size:
      Long): boolean; override;

    {This function converts png data to ImageData structure.}
    function DecodeBuffer(Data: Pointer; Size: Long; var Image: TImageData):
      boolean; override;

    {This func takes valid png image data and converts it to specified image
    file format ready to be written into file.}
    function EncodeStream(const Image: TImageData; Output: TStream): boolean;
      override;

    {This func converts png in Input stream read from image file to Image Data
    structure.}
    function DecodeStream(Input: TStream; var Image: TImageData): boolean;
      override;

    {This function tests if data are valid png image.}
    function TestFormat(Data: Pointer; Size: Long): boolean; override;

    {This function tests if stream contains valid png.}
    function TestFormat(Input: TStream): boolean; override;

  end;


type

  {Chunk header}
  TChunkHdr = packed record
    DataSize: ULong;
    ChunkID: ULong;
  end;

  {IHDR chunk format}
  TIHDR = packed record
    ImgWidth: ULong;
    ImgHeight: ULong;
    BitDepth: Byte; // bits per pixel  or bits per sample (for truecolor)
    ColorType: Byte; // 0 = greyscale, 2 = truecolor, 3 = palette,
                    // 4 = grey+alpha, 6 = truecolor + alpha
    Compress: Byte; // compression type  00 = ZLib
    Filter: Byte; // used precompress filter
    Interlaced: Byte; // 0 = no int; 1 = Adam7 interlacing
  end;


const


  {Constants for chunk identifiers and signature identifiers.
  They are already in big-endian format.}
  IHDRChunk = $49484452;
  GAMAChunk = $67414D41;
  PLTEChunk = $504C5445;
  BKGDChunk = $624B4744;
  TRNSChunk = $74524E53;
  IDATChunk = $49444154;
  IENDChunk = $49454E44;
  CheckSig: array[0..7] of Byte = ($89, $50, $4E, $47, $0D, $0A, $1A, $0A);

  { Interlace start and offsets}
  RowStart: array[0..6] of Long = (0, 0, 4, 0, 2, 0, 1);
  ColumnStart: array[0..6] of Long = (0, 4, 0, 2, 0, 1, 0);
  RowIncrement: array[0..6] of Long = (8, 8, 8, 4, 4, 2, 2);
  ColumnIncrement: array[0..6] of Long = (8, 8, 4, 4, 2, 2, 1);
  PassMask: array[0..6] of Byte = ($80, $08, $88, $22, $AA, $55, $FF);


  sPngName = 'Png Module';
  sPngDesc =
    'Provides Portable Network Graphics (*.png) image format encoding/decoding';
  sPngCode: TIDCode = 'IMG-PNG';


var

  PngModuleInfo:TModuleInfo;


implementation


{*************************************************************************}
{                         Some support functions                          }
{*************************************************************************}


function PaethPredictor(a, b, c: Byte): Byte;
var
  p, pa, pb, pc: Integer;

begin
  p := a + b - c;
  pa := Abs(p - a);
  pb := Abs(p - b);
  pc := Abs(p - c);
  if (pa <= pb) and (pa <= pc) then
    Result := a
  else if pb <= pc then
    Result := b
  else
    Result := c;
end;

procedure Convert1To8(var Image: TImageData);
const
  Mask1: array[0..7] of Byte = ($80, $40, $20, $10, $08, $04, $02, $01);
  Shift1: array[0..7] of Byte = (7, 6, 5, 4, 3, 2, 1, 0);

var
  Data: Pointer;
  X, Y: Long;
  WB: Long;
begin
  with Image, Image.Levels[0] do
  begin
    if (BitCount <> 1) or (Bits = nil) then
      Exit;
    Size := Width * Height;
    GetMem(Data, Size);
    WB := Width div 8;
    if Odd(Size) then
      Inc(WB);
    try
      if PalSize > 0 then
      begin
        PalSize := 256;
        ReallocMem(Palette, PalSize * SizeOf(TARGBRec));
      end;
      for X := 0 to Width - 1 do
        for Y := 0 to Height - 1 do
          PByte(Long(Data) + Y * Width + X)^ :=
            (PByteArray(Integer(Bits) + Y * WB)[X shr 3] and Mask1[X and 7]) shr
            Shift1[X and 7];
    finally
      FreeMemNil(Bits);
      Bits := Data;
      if PalSize > 0 then
        ImageFormat := ifIndex8
      else
        ImageFormat := ifIntensity8;
      BitCount := 8;
    end;
  end;
end;

procedure Convert4To8(var Image: TImageData);
const
  Mask4: array[0..1] of Byte = ($F0, $0F);
  Shift4: array[0..1] of Byte = (4, 0);

var
  Data: Pointer;
  X, Y: Long;
  WB: Long;
begin
  with Image, Image.Levels[0] do
  begin
    if (BitCount <> 4) or (Bits = nil) then
      Exit;

    Size := Width * Height;
    GetMem(Data, Size);
    WB := Width div 2;
    if Odd(Size) then
      Inc(WB);
    try
      if PalSize > 0 then
      begin
        PalSize := 256;
        ReallocMem(Palette, PalSize * SizeOf(TARGBRec));
      end;
      for X := 0 to Width - 1 do
        for Y := 0 to Height - 1 do
          PByte(Long(Data) + Y * Width + X)^ :=
            (PByteArray(Integer(Bits) + Y * WB)[X shr 1] and Mask4[X and 1]) shr
            Shift4[X and 1];
    finally
      FreeMemNil(Bits);
      Bits := Data;
      if PalSize > 0 then
        ImageFormat := ifIndex8
      else
        ImageFormat := ifIntensity8;
      BitCount := 8;
    end;
  end;
end;


{*************************************************************************}
{                         TPngModule Methods                              }
{*************************************************************************}


constructor TPngModule.Create;
begin
  inherited Create;
  FInfo := PngModuleInfo;
end;

destructor TPngModule.Destroy;
begin

  inherited Destroy;
end;

procedure TPngModule.ApplyFilter(Filter: Byte; Line, PrevLine, Target: PByte;
  BPP, BytesPerRow: Integer);
var
  I: Long;
  Raw,
    Decoded,
    Prior,
    PriorDecoded,
    TargetRun: PByte;

begin
  case Filter of
    0: // no filter, just copy data
      Move(Line^, Target^, BytesPerRow);
    1: // subtraction filter
      begin
        Raw := Line;
        TargetRun := Target;
        // Transfer BPP bytes without filtering. This mimics the effect of bytes left to the
        // scanline being zero.
        Move(Raw^, TargetRun^, BPP);

        // now do rest of the line
        Decoded := TargetRun;
        Inc(Raw, BPP);
        Inc(TargetRun, BPP);
        Dec(BytesPerRow, BPP);
        while BytesPerRow > 0 do
        begin
          TargetRun^ := Byte(Raw^ + Decoded^);
          Inc(Raw);
          Inc(Decoded);
          Inc(TargetRun);
          Dec(BytesPerRow);
        end;
      end;
    2: // Up filter
      begin
        Raw := Line;
        Prior := PrevLine;
        TargetRun := Target;
        while BytesPerRow > 0 do
        begin
          TargetRun^ := Byte(Raw^ + Prior^);
          Inc(Raw);
          Inc(Prior);
          Inc(TargetRun);
          Dec(BytesPerRow);
        end;
      end;
    3: // average filter
      begin
        // first handle BPP virtual pixels to the left
        Raw := Line;
        Decoded := Line;
        Prior := PrevLine;
        TargetRun := Target;
        for I := 0 to BPP - 1 do
        begin
          TargetRun^ := Byte(Raw^ + Floor(Prior^ / 2));
          Inc(Raw);
          Inc(Prior);
          Inc(TargetRun);
        end;
        Dec(BytesPerRow, BPP);

        // now do rest of line
        while BytesPerRow > 0 do
        begin
          TargetRun^ := Byte(Raw^ + Floor((Decoded^ + Prior^) / 2));
          Inc(Raw);
          Inc(Decoded);
          Inc(Prior);
          Inc(TargetRun);
          Dec(BytesPerRow);
        end;
      end;
    4: // paeth prediction
      begin
       // again, start with first BPP pixel which would refer to non-existing pixels to the left
        Raw := Line;
        Decoded := Target;
        Prior := PrevLine;
        PriorDecoded := PrevLine;
        TargetRun := Target;
        for I := 0 to BPP - 1 do
        begin
          TargetRun^ := Byte(Raw^ + PaethPredictor(0, Prior^, 0));
          Inc(Raw);
          Inc(Prior);
          Inc(TargetRun);
        end;
        Dec(BytesPerRow, BPP);

        // finally do rest of line
        while BytesPerRow > 0 do
        begin
          TargetRun^ := Byte(Raw^ + PaethPredictor(Decoded^, Prior^,
            PriorDecoded^));
          Inc(Raw);
          Inc(Decoded);
          Inc(Prior);
          Inc(PriorDecoded);
          Inc(TargetRun);
          Dec(BytesPerRow);
        end;
      end;
  end;
end;

procedure TPngModule.PrepareImage(const Image: TImageData; out Clone:
  TImageData);
begin
  inherited PrepareImage(Image, Clone);
  with Clone do
  begin
    // PNG does not support 8/16bit RGB images -> convert to 24/32bit
    case BitCount of
     16:
      if ImageFormatToPF(ImageFormat).ABitCount> 0 then
          Convert16To32(Clone)
       else
          Convert16To24(Clone);
      8:
      if ImageFormatToPF(ImageFormat).ABitCount> 0 then
          Convert8To32(Clone)
       else
          Convert8To24(Clone);
     end;

    if BitCount in [24, 32] then
      SwapChannels(Clone);
  end;
end;

function TPngModule.DecodeBuffer(Data: Pointer; Size: Long;
  var Image: TImageData): boolean;
var
  MS: TMemoryStream;

begin
  MS := TMemoryStream.Create;
  MS.WriteBuffer(Data^, Size);
  // this is important line
  MS.Position := 0;
  try
    // calls DecodeStream function rather that try to parse raw pointer
    Result := DecodeStream(MS, Image);
  finally
    MS.Free;
  end;
end;

function TPngModule.DecodeStream(Input: TStream;
  var Image: TImageData): boolean;
var
  OldPos, ID: Int64;
  Buff: Pointer;
  ChunkHdr: TChunkHdr;
  IHDR: TIHDR;
  CRC: ULong;
  FRawData: Pointer;
  FCompData: Pointer;


  procedure DecodeInit;
  begin
    FillChar(FStream, sizeof(FStream), 0);
    FStream.zalloc := zcAlloc;
    FStream.zfree := zcFree;
    InflateInit_(FStream, zlib_version, sizeof(FStream));
  end;


  procedure Decode(var InData, OutData: Pointer; PackedSize,
    UnpackedSize: Long);
  begin
    FStream.next_in := InData;
    FStream.avail_in := PackedSize;
    FStream.next_out := OutData;
    FStream.avail_out := UnpackedSize;
    Inflate(FStream, Z_PARTIAL_FLUSH);
    // advance pointers so used input can be calculated
    InData := FStream.Next_In;
    OutData := FStream.Next_Out;
  end;


  procedure DecodeEnd;
  begin
    InflateEnd(FStream);
  end;


  procedure ReadRow(RowBuffer: Pointer; BytesPerRow: Integer);
  var
    LocalBuffer: Pointer;
    PendingOutput: Integer;
    OPos: Long;

  begin
    LocalBuffer := RowBuffer;
    PendingOutput := BytesPerRow;

    repeat
      if ChunkHdr.DataSize + Long(FRawData) = Long(FCompData) then
      begin
          // read next IDAT chunk
        OPos := Input.Position;
        Input.ReadBuffer(CRC, 4);
        Input.ReadBuffer(ChunkHdr, SizeOf(ChunkHdr));
        SwapLong(@ChunkHdr, 2);
        if ChunkHdr.ChunkID = IDATChunk then
        begin
          FreeMemNil(FRawData);
          GetMem(FRawData, ChunkHdr.DataSize);
          Input.ReadBuffer(FRawData^, ChunkHdr.DataSize);
          FCompData := FRawData;
        end
        else
        begin
          Input.Position := OPos;
          Exit;
        end;
      end;

      // this decode call will advance InData and OutData accordingly
      Decode(FCompData,
        LocalBuffer,
        ChunkHdr.DataSize - (Long(FCompData) - Long(FRawData)),
        PendingOutput);

      PendingOutput := BytesPerRow - (Long(LocalBuffer) - Long(RowBuffer));
    until PendingOutput = 0;
  end;


  procedure LoadIDAT;
  var
    I: Long;
    RowBuffer: array[Boolean] of PChar;
    EvenRow: boolean;
    Pass: Long;
    BytesPerRow,
      InterlaceRowBytes,
      InterlaceWidth: Long;


    procedure DecodeAdam7;
    const
      BitTable: array[1..8] of Integer = ($1, $3, 0, $F, 0, 0, 0, $FF);
      StartBit: array[1..8] of Integer = (7, 0, 0, 4, 0, 0, 0, 0);
    var
      Dst2: PChar;
      Src, Dst: PByte;
      CurBit, Col: Long;
    begin
      Src := PByte(Long(RowBuffer[EvenRow]) + 1);
      Col := ColumnStart[Pass];
      with Image do
        case BitCount of
          1, 4, 8:
          begin
          Dst := Pointer(Long(Image.Levels[0].Bits) + I * (BytesPerRow - 1));
          repeat
              CurBit := StartBit[BitCount];
              repeat
                Dst2 := PChar(Long(Dst) +(BitCount * Col) shr 3);
                Byte(Dst2^) := Byte(Dst2^) or
                  (((Byte(Src^) shr CurBit) and BitTable[BitCount])
                  shl (StartBit[BitCount] - (Col * BitCount mod 8)));

                Inc(Col, ColumnIncrement[Pass]);
                Dec(CurBit, BitCount);
              until CurBit < 0;
              Inc(Src);
            until Col >= Levels[0].Width;
            end;
          16:
            begin
              Dst := Pointer(Long(Image.Levels[0].Bits) + I * (BytesPerRow - 1)
                + Col*2);
              repeat
                Byte(Dst^) := PByte(Long(Src) + 0)^;
                Inc(Dst);
                Byte(Dst^) := PByte(Long(Src) + 1)^;
                Inc(Dst);

                Inc(Src, 2);
                Inc(Dst, ColumnIncrement[Pass] *2-2);
                Inc(Col, ColumnIncrement[Pass]);
              until Col >= Levels[0].Width;
            end;
          24:
            begin
              Dst := Pointer(Long(Image.Levels[0].Bits) + I * (BytesPerRow - 1)
                + Col * 3);
              repeat
                Byte(Dst^) := PByte(Long(Src) + 0)^;
                Inc(Dst);
                Byte(Dst^) := PByte(Long(Src) + 1)^;
                Inc(Dst);
                Byte(Dst^) := PByte(Long(Src) + 2)^;
                Inc(Dst);

                Inc(Src, 3);
                Inc(Dst, ColumnIncrement[Pass] * 3 - 3);
                Inc(Col, ColumnIncrement[Pass]);
              until Col >= Levels[0].Width;
            end;
          32:
            begin
              Dst := Pointer(Long(Image.Levels[0].Bits) + I * (BytesPerRow - 1)
                + Col * 4);
              repeat

                Byte(Dst^) := PByte(Long(Src) + 0)^;
                Inc(Dst);
                Byte(Dst^) := PByte(Long(Src) + 1)^;
                Inc(Dst);
                Byte(Dst^) := PByte(Long(Src) + 2)^;
                Inc(Dst);
                Byte(Dst^) := PByte(Long(Src) + 3)^;
                Inc(Dst);

                Inc(Src, 4);
                Inc(Dst, ColumnIncrement[Pass] * 4 - 4);
                Inc(Col, ColumnIncrement[Pass]);
              until Col >= Levels[0].Width;
            end;
        end;
    end;


  begin
    RowBuffer[False] := nil;
    RowBuffer[True] := nil;
    with Image, Image.Levels[0] do
    try

      // maximum number of bytes per row and consider there's one filter byte at the start of each row
      // +7 is here fot 1/4 bit images support
      BytesPerRow := ((Width * BitCount + 7) shr 3) + 1;

      GetMem(FRawData, ChunkHdr.DataSize);
      Input.ReadBuffer(FRawData^, ChunkHdr.DataSize);
      FCompData := FRawData;

      Size := Height * (BytesPerRow - 1);
      GetMem(Bits, Size);

      GetMem(RowBuffer[true], BytesPerRow);
      GetMem(RowBuffer[false], BytesPerRow);

      // there can be more than one IDAT chunk in the file but then they must directly
      // follow each other (handled in ReadRow)
      EvenRow := True;

      DecodeInit;

      // prepare interlaced images
      if IHDR.Interlaced = 1 then
      begin
        for Pass := 0 to 6 do
        begin
          // prepare next interlace run
          if Width <= ColumnStart[Pass] then
            Continue;
          InterlaceWidth := (Width + ColumnIncrement[Pass] - 1 -
            ColumnStart[Pass])
            div ColumnIncrement[Pass];
          InterlaceRowBytes := ((InterlaceWidth * BitCount + 7) shr 3) + 1;

          I := RowStart[Pass];
          while I < Height do
          begin
            ReadRow(RowBuffer[EvenRow], InterlaceRowBytes);
            ApplyFilter(Byte(RowBuffer[EvenRow]^),
              Pointer(RowBuffer[EvenRow] + 1),
              Pointer(RowBuffer[not EvenRow] + 1),
              Pointer(RowBuffer[EvenRow] + 1),
              (BitCount + 7) shr 3,
              InterlaceRowBytes - 1);

            DecodeAdam7;

            EvenRow := not EvenRow;
            // continue with next row in interlaced order
            Inc(I, RowIncrement[Pass]);
          end;
        end;
      end
      else
      begin
        for I := 0 to Height - 1 do
        begin
          ReadRow(RowBuffer[EvenRow], BytesPerRow);
          ApplyFilter(Byte(RowBuffer[EvenRow]^),
            Pointer(RowBuffer[EvenRow] + 1),
            Pointer(RowBuffer[not EvenRow] + 1),
            Pointer(RowBuffer[EvenRow] + 1),
            (BitCount + 7) shr 3,
            BytesPerRow - 1);

          Move(Pointer(RowBuffer[EvenRow] + 1)^,
            Pointer(Long(Bits) + I * (BytesPerRow - 1))^, BytesPerRow - 1);

          EvenRow := not EvenRow;
        end;
      end;

    finally
      FreeMemNil(FRawData);
      FreeMemNil(RowBuffer[True]);
      FreeMemNil(RowBuffer[False]);
      DecodeEnd;
    end;
  end;


  procedure LoadPLTE;
  var
    Pal: PPalRGBArray;

    function ConvPal(Pal: PPalRGBArray; Entries: Long): PPalArray;
    var
      I: Long;
    begin
      GetMemFill(Result, Entries * SizeOf(TARGB));
      for I := 0 to Entries - 1 do
      begin
        Result[I].A := $FF;
        Result[I].R := Pal[I].R;
        Result[I].G := Pal[I].G;
        Result[I].B := Pal[I].B;
      end;
    end;

  begin
    if IHDR.ColorType <> 3 then
      Exit;

    with Image do
    try
      GetMemFill(Pal, ChunkHdr.DataSize);
      Input.ReadBuffer(Pal^, ChunkHdr.DataSize);
      PalSize := ChunkHdr.DataSize div 3;

      Palette := ConvPal(Pal, PalSize);

    finally
      FreeMemNil(Pal);
    end;
  end;


begin
  Result := false;
  FreeImageSafe(Image);
  Assert(TestFormat(Input), ErrString(Self, 'DecodeStream',
    'Buffer image data not in Png format or corrupted!'));

  OldPos := Input.Position;

  AddLevel(Image);

  with Image, Image.Levels[0] do
  try
    try
      // read signature
      Input.ReadBuffer(ID, 8);
      // read IHDR
      Input.ReadBuffer(ChunkHdr, SizeOf(ChunkHdr));
      Input.ReadBuffer(IHDR, SizeOf(IHDR));
      Input.ReadBuffer(CRC, SizeOf(CRC));
      // swap width and height, next values are single bytes
      SwapLong(@IHDR, 2);
      Width := IHDR.ImgWidth;
      Height := IHDR.ImgHeight;

      case IHDR.ColorType of
        0: case IHDR.BitDepth of
            1..8:
              begin
                BitCount := IHDR.BitDepth;
                ImageFormat := ifIntensity8;
              end;
            16:
              begin
                BitCount := 16;
                ImageFormat := ifIntensity16;
              end;
          end;
        2: case IHDR.BitDepth of
            8:
              begin
                BitCount := 24;
                ImageFormat := ifR8G8B8;
              end;
          end;
        3:
          begin
            case IHDR.BitDepth of
              1: BitCount := 1;
              4: BitCount := 4;
              8: BitCount := 8;
            end;
            // all indexed modes will be converted to 8bit
            ImageFormat := ifIndex8;
          end;
        // greyscale + alpha hybrid format is not supported -> treat like intensity format
        4:
          begin
            BitCount := IHDR.BitDepth * 2;
            case IHDR.BitDepth of
              1..4: ImageFormat := ifIntensity8;
              8: ImageFormat := ifIntensity16;
              16: ImageFormat := ifIntensity32;
            end;
          end;
        6: case IHDR.BitDepth of
            8:
              begin
                BitCount := 32;
                ImageFormat := ifA8R8G8B8;
              end;
          end;
      end;

      Input.ReadBuffer(ChunkHdr, SizeOf(ChunkHdr));
      SwapLong(@ChunkHdr, 2);
      // read chunks until IEND chunk is reached
      while ChunkHdr.ChunkID <> IENDCHUNK do
      begin
        case ChunkHdr.ChunkID of
          IDATChunk: LoadIDAT;
          PLTEChunk: LoadPLTE;
        else
          // skip unsupported chunks
          if Input.Size - Input.Position > ChunkHdr.DataSize then
          begin
            GetMem(Buff, ChunkHdr.DataSize);
            Input.ReadBuffer(Buff^, ChunkHdr.DataSize);
            FreeMemNil(Buff);
          end
          else
            // some programs do not add IEND !! (ACDSee)
            Break;
        end;
        Input.Read(CRC, SizeOf(CRC));
        Input.Read(ChunkHdr, SizeOf(ChunkHdr));
        SwapLong(@ChunkHdr, 2);
      end;

      case BitCount of
        1: Convert1To8(Image);
        4: Convert4To8(Image);
      end;

      SwapChannels(Image, 0);

    except
      RaiseError(ErrString(Self, 'DecodeStream',
        'Error while loading Png data! Image can be corrupted.'),EImage);
      Input.Position := OldPos;
    end;
  finally

  end;
  Result := true;
end;

function TPngModule.EncodeBuffer(const Image: TImageData;
  out Data: Pointer; out Size: Long): boolean;
var
  MS: TMemoryStream;

begin
  MS := TMemoryStream.Create;
  try
    // calls EncodeStream function rather that try to parse raw pointer
    Result := EncodeStream(Image, MS);
  finally
    Size := MS.Size;
    GetMem(Data, Size);
    // this is important line
    MS.Position := 0;
    MS.ReadBuffer(Data^, Size);
    MS.Free;
  end;
end;

function TPngModule.EncodeStream(const Image: TImageData;
  Output: TStream): boolean;
var
  OldPos: Int64;
  ChunkHdr: TChunkHdr;
  IHDR: TIHDR;
  CRC: ULong;
  Clone: TImageData;


  procedure SaveIDAT;
  var
    BytesPerRow,
      OutSize,
      OPos, OPos2,
      I: Long;
    Buff, InData,
      OutData: Pointer;

  begin
    with Clone, Clone.Levels[0] do
    try
      BytesPerRow := Width * (BitCount shr 3);

      OPos := OutPut.Position;
      ChunkHdr.DataSize := 0;
      ChunkHdr.ChunkID := SwapLong(IDATChunk);
      Output.WriteBuffer(ChunkHdr, SizeOf(ChunkHdr));

      CRC := $FFFFFFFF;
      CalcCRC32(Pointer(Long(@ChunkHdr) + 4), 4, CRC);

      GetMem(InData, BytesPerRow + 1);
      GetMem(Buff, Size + Height);

      for I := 0 to Height - 1 do
      begin
        Move(Pointer(Long(Bits) + I * BytesPerRow)^, Pointer(Long(InData)
          + 1)^,
          BytesPerRow);
        PByte(InData)^ := 0;
        Move(InData^, Pointer(Long(Buff) + I * (BytesPerRow + 1))^,
          BytesPerRow
          + 1);
      end;

      CompressBuffer(Buff, Size + Height, OutData, OutSize);
      Output.WriteBuffer(OutData^, OutSize);

      OPos2 := OutPut.Position;
      ChunkHdr.DataSize := SwapLong(OPos2 - Opos - SizeOf(ChunkHdr));
      OutPut.Position := OPos;
      Output.WriteBuffer(ChunkHdr, SizeOf(ChunkHdr));
      OutPut.Position := OPos2;

      CalcCRC32(OutData, OutSize, CRC);
      CRC := SwapLong(CRC xor $FFFFFFFF);
      Output.WriteBuffer(CRC, 4);

    finally
      FreeMemNil(InData);
      FreeMemNil(OutData);
      FreeMemNil(Buff);
    end;
  end;


  procedure SavePLTE;
  var
    Pal: TPalRGB;
    I: Long;

  begin
    with Clone do
      for I := 0 to PalSize - 1 do
      begin
        Pal[I].B := Palette[I].R;
        Pal[I].G := Palette[I].G;
        Pal[I].R := Palette[I].B;
      end;
    ChunkHdr.ChunkID := PLTEChunk;
    ChunkHdr.DataSize := SizeOf(Pal);
    SwapLong(@ChunkHdr, 2);
    Output.WriteBuffer(ChunkHdr, SizeOf(ChunkHdr));
    Output.WriteBuffer(Pal, SizeOf(Pal));

    CRC := $FFFFFFFF;
    CalcCRC32(Pointer(Long(@ChunkHdr) + 4), 4, CRC);
    CalcCRC32(@Pal, SizeOf(Pal), CRC);
    CRC := SwapLong(CRC xor $FFFFFFFF);
    Output.WriteBuffer(CRC, 4);
  end;


begin
  Result := false;
  Assert(TestImage(Image), ErrString(Self, 'EncodeStream',
    'Image data information is invalid! Image parameter: ' + #10#13 +
    ImageToStr(Image)));

  PrepareImage(Image, Clone);

  OldPos:=Output.Position;

  with Clone, Clone.Levels[0] do
  try
    try
      Output.WriteBuffer(CheckSig, 8);
      ChunkHdr.DataSize := SizeOf(IHDR);
      ChunkHdr.ChunkID := IHDRChunk;
      SwapLong(@ChunkHdr, 2);
      IHDR.ImgWidth := Width;
      IHDR.ImgHeight := Height;
      IHDR.Compress := 0;
      IHDR.Filter := 0;
      IHDR.Interlaced := 0;

      case ImageFormat of
        ifIntensity8:
          begin
            IHDR.BitDepth := 8;
            IHDR.ColorType := 0;
          end;
        ifIntensity16:
          begin
            IHDR.BitDepth := 16;
            IHDR.ColorType := 0;
          end;
        ifIndex8:
          begin
            IHDR.BitDepth := 8;
            IHDR.ColorType := 3;
          end;
        ifR8G8B8:
          begin
            IHDR.BitDepth := 8;
            IHDR.ColorType := 2;
          end;
        ifA8R8G8B8:
          begin
            IHDR.BitDepth := 8;
            IHDR.ColorType := 6;
          end;
      end;

      SwapLong(@IHDR, 2);
      Output.WriteBuffer(ChunkHdr, SizeOf(ChunkHdr));
      Output.WriteBuffer(IHDR, SizeOf(IHDR));
      CRC := $FFFFFFFF;
      CalcCRC32(Pointer(Long(@ChunkHdr) + 4), 4, CRC);
      CalcCRC32(@IHDR, SizeOf(IHDR), CRC);
      CRC := SwapLong(CRC xor $FFFFFFFF);
      Output.WriteBuffer(CRC, 4);

      if IHDR.ColorType = 3 then
        SavePLTE;

      SaveIDAT;

      ChunkHdr.DataSize := 0;
      ChunkHdr.ChunkID := SwapLong(IENDChunk);
      CRC := $FFFFFFFF;
      CalcCRC32(Pointer(Long(@ChunkHdr) + 4), 4, CRC);
      CRC := SwapLong(CRC xor $FFFFFFFF);
      Output.WriteBuffer(ChunkHdr, SizeOf(ChunkHdr));
      Output.WriteBuffer(CRC, 4);

    except
      RaiseError(ErrString(Self, 'EncodeStream',
        'Error while saving Png data! Image can be corrupted.'),EImage);
      OutPut.Position := OldPos;
    end;
  finally
    FreeImageSafe(Clone);
  end;
  Result := true;
end;

function TPngModule.TestFormat(Input: TStream): boolean;
var
  OldPos: Int64;
  C: array[0..7]of Char;
begin
  Result := false;
  if Input = nil then
    Exit;

  OldPos := Input.Position;
  Input.ReadBuffer(C, SizeOf(C));
  Input.Position := OldPos;

  if CompareMem(@C, @CheckSig, SizeOf(C)) then
    Result := true;
end;

function TPngModule.TestFormat(Data: Pointer; Size: Long): boolean;
begin
  Result := false;
  if (Data = nil) or (Size = 0) then
    Exit;

  if CompareMem(PULong(Data), @CheckSig, SizeOf(CheckSig)) then
    Result := true;
end;


initialization

  PngModuleInfo:=GetDefaultIntModuleInfo;
  PngModuleInfo.Name:=sPngName;
  PngModuleInfo.Description:=sPngDesc;
  PngModuleInfo.IDCode:=sPngCode;
  RegisterIntModule(@PngModuleInfo,TPngModule);

end.

